# -*- encoding: utf-8 -*-
import sys
from collections import namedtuple
import multiprocessing
import logging
import kamma

try:
    MAX_WAIT = sys.maxint / 2
except AttributeError:
    MAX_WAIT = 1073741823


logger = logging.getLogger(__name__)

""" task to be done."""
Task = namedtuple('Task', ['callback', 'kwargs'])
exec_info = namedtuple('exec_info', ['attempts', 'delay'])


class stop_none(object):
    """Non stop strategy."""

    def __call__(self, previous_attempt_number, delay_since_first_attempt):
        return False


class stop_after_attempt(object):
    """Stop strategy that stops when the previous attempt >= max_attempt."""

    def __init__(self, max_attempt_number):
        self.max_attempt_number = max_attempt_number

    def __call__(self, previous_attempt_number, delay_since_first_attempt):
        return previous_attempt_number >= self.max_attempt_number


class stop_after_delay(object):
    """Stop strategy that stops when the time from the first attempt >= limit."""

    def __init__(self, max_delay):
        self.max_delay = max_delay

    def __call__(self, previous_attempt_number, delay_since_first_attempt):
        return delay_since_first_attempt >= self.max_delay


class wait_fixed(object):
    """Wait strategy that waits a fixed amount of time between each retry."""

    def __init__(self, wait):
        self.wait_fixed = wait

    def __call__(self, previous_attempt_number, delay_since_first_attempt):
        return self.wait_fixed


class wait_incremental(object):
    """Wait an incremental amount of time after each attempt.
    Starting at a starting value and incrementing by a value for each attempt
    (and restricting the upper limit to some maximum value).
    """

    def __init__(self, start=0, increment=100, max=MAX_WAIT):
        self.start = start
        self.increment = increment
        self.max = max

    def __call__(self, previous_attempt_number, delay_since_first_attempt):
        result = self.start + (self.increment * previous_attempt_number)
        return max(0, min(result, self.max))


class wait_exponential(object):
    """Wait strategy that applies exponential backoff.
    It allows for a customized multiplier and an ability to restrict the
    upper limit to some maximum value.
    """

    def __init__(self, exp_base=2, multiplier=1, max=MAX_WAIT):
        self.multiplier = multiplier
        self.exp_base = exp_base
        self.max = max

    def __call__(self, previous_attempt_number, delay_since_first_attempt):
        try:
            exp = self.exp_base ** previous_attempt_number
            result = self.multiplier * exp
        except OverflowError:
            return self.max
        return max(0, min(result, self.max))


# task is a piece of work to be done
class TaskCallback(object):
    """Task definition by it 'id', callback, timeout in seconds, retry wait and retry stop.
    WARNING: callbacks cannot be nested functions (python in windows)
    """

    def __init__(self, callback, timeout, retry_wait, retry_stop):
        self._callback = callback
        self._timeout = timeout
        self._retry_wait = retry_wait
        self._retry_stop = retry_stop

    def execute(self, quit_event, **kwargs):
        previous_attempt_number = 0
        delay_since_first_attempt = 0
        last_error = ""
        while not self._retry_stop(previous_attempt_number, delay_since_first_attempt):
            logger.debug("Executing task '{cb}' try: {attempt} total wait: {delay}".format(cb=self._callback.__name__,
                                                                                           attempt=previous_attempt_number + 1,
                                                                                           delay=delay_since_first_attempt))
            try:
                if self._execute(**kwargs):
                    return exec_info(attempts=previous_attempt_number + 1, delay=delay_since_first_attempt)
            except kamma.AbortTask as e:
                raise e
            except Exception as e:
                last_error = str(e)
                logger.warning("Exception running task '{cb}': {e}".format(cb=self._callback.__name__, e=last_error))
            wait = self._retry_wait(previous_attempt_number, delay_since_first_attempt)
            quit_event.wait(wait)
            previous_attempt_number += 1
            delay_since_first_attempt += wait
        raise kamma.RetryStopped(callback=self._callback.__name__,
                                 attempts=previous_attempt_number,
                                 delay=delay_since_first_attempt,
                                 last_error=last_error)

    def _execute(self, **kwargs):
        # Start bar as a process
        # concurrent.futures were introduced in python 3.2
        exception_queue = multiprocessing.Queue()
        p = multiprocessing.Process(target=self._run_callback, args=(self._callback, exception_queue,), kwargs=kwargs)
        # p.daemon = True
        p.start()

        # Wait for self._timoeut seconds or until process finishes
        p.join(self._timeout)

        # If thread is still active
        if p.is_alive():
            # Terminate
            p.terminate()
            p.join()
            raise kamma.TimeoutTask("Timeout triggered after {sec} seconds.".format(sec=self._timeout))
        # process child exception
        if not exception_queue.empty():
            raise exception_queue.get_nowait()
        logger.debug("Task '{cb}' executed successfully".format(cb=self._callback.__name__))
        return True

    def _run_callback(self, callback, exception_queue, **kwargs):
        try:
            callback(**kwargs)
        except Exception as e:
            exception_queue.put(e)



# -*- encoding: utf-8 -*-
import sys
import os
from collections import namedtuple
import threading
import traceback
import logging
import json

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import kamma
from kamma import task
from kamma.queue import FileQueue


_task = namedtuple('_task', ['callback', 'retry'])

logger = logging.getLogger(__name__)


class Kamma(object):
    def __init__(self, queue_path="task_queue"):
        self._quit = False
        self._thread = None
        self._exception = None
        self._on_failure = None
        self._on_abortion = None
        self._callbacks = dict()
        self._mutex = threading.Lock()
        self._empty_event = threading.Event()
        self._push_event = threading.Event()
        self._quit_event = threading.Event()
        self._queue = FileQueue(queue_path)

    def __del__(self):
        self.stop()
        self._callbacks = None
        self._queue = None

    def task_callback(self, timeout=(4 * 60 * 60), retry_wait=task.wait_fixed(15), retry_stop=task.stop_none()):
        ''' Registers a Task callback in Kamma.
        '''
        def decorator(func):
            self.add_task_callback(callback=func, timeout=timeout, retry_wait=retry_wait, retry_stop=retry_stop)
            return func
        return decorator

    def on_failure(self):
        ''' Registers a on_failure callback in Kamma. It will be called each time that error occurs
        '''
        def decorator(func):
            self.add_on_failure(func)
            return func
        return decorator

    def on_abortion(self):
        ''' Registers a on_anortion callback in Kamma. It will be called each time that task is aborted
        '''
        def decorator(func):
            self.add_on_abortion(func)
            return func
        return decorator

    def add_task_callback(self, callback, timeout=(4 * 60 * 60), retry_wait=task.wait_fixed(15), retry_stop=task.stop_none()):
        ''' Registers a Task in Kamma.
        :param function callback: The name of the callback function has to be unique.
        '''
        if callback.__name__ in self._callbacks:
            raise kamma.TaskAlreadyRegistered('the task \'{}\' is already used'.format(callback.__name__))
        self._callbacks[callback.__name__] = task.TaskCallback(callback, timeout=timeout, retry_wait=retry_wait, retry_stop=retry_stop)

    def add_on_failure(self, callback):
        self._on_failure = callback

    def add_on_abortion(self, callback):
        self._on_abortion = callback

    def push_task(self, callback, **kwargs):
        logger.debug('push task \'{}\' with kwargs: {}'.format(callback.__name__, kwargs))
        with self._mutex:
            if callback.__name__ not in self._callbacks:
                raise kamma.TaskNotRegistered('the task \'{}\' is not registered'.format(callback.__name__))
            self._queue.push(json.dumps({'callback': callback.__name__, 'kwargs': kwargs}))
            self._empty_event.clear()
            self._push_event.set()

    def stop(self):
        with self._mutex:
            self._quit = True
            self._push_event.set()
            self._quit_event.set()
            if self._thread and self._thread.is_alive():
                logger.info('stopping')
                self._thread.join()
                logger.info('stopped')

    def pending(self):
        return self._queue.length()

    def wait_empty_event(self, timeout=None):
        logger.info('waiting')
        if self._queue.length() == 0:
            self._empty_event.set()
        self._empty_event.wait(timeout)
        if self._exception:
            e = self._exception
            self._exception = None
            raise e

    def _get_task_callback(self, callback):
        if callback not in self._callbacks:
            raise kamma.TaskNotRegistered('the task \'{}\' is not registered'.format(callback))
        return self._callbacks.get(callback, None)

    def run(self):
        logger.info("running")
        try:
            while not self._quit:
                if self._queue.length() > 0:
                    self._process_queue()
                self._push_event.wait(10)
        except Exception as e:
            logger.error(traceback.format_exc())
            with self._mutex:
                self._exception = e
                self._empty_event.set()
                raise self._exception
        finally:
            logger.info("exiting")

    def run_async(self):
        self._thread = threading.Thread(target=self.run, args=())
        self._thread.start()

    def _process_queue(self):
        count = 0
        while not self._quit and self._queue.length() > 0:
            json_task = self._queue.head()
            task_queued = json.loads(json_task)
            task_callback = self._get_task_callback(task_queued['callback'])
            task_kwargs = task_queued['kwargs']
            try:
                task_callback.execute(quit_event=self._quit_event, **task_kwargs)
                self._queue.pop()
            except kamma.AbortTask as e:
                logger.warning('AbortTask exception from \'{}\' due to: {}'.format(task_queued['callback'], e))
                if self._on_abortion:
                    self._on_abortion(json_task=json_task, reason=str(e))
                self._queue.pop()
            except kamma.RetryStopped as e:
                logger.warning('RetryStopped exception from \'{}\' due to: {}'.format(task_queued['callback'], e))
                if self._on_failure:
                    self._on_failure(json_task=json_task, retry_stopped=e)
                self._queue.pop()
            count = count + 1
        if count > 0:
            logger.info("processed {} queued tasks".format(count))
        with self._mutex:
            if self._queue.length() == 0:
                self._push_event.clear()
                self._empty_event.set()
                logger.debug("set empty event")


if __name__ == "__main__":
    worker = Kamma(tasks=dict())
    worker.stop()



# -*- encoding: utf-8 -*-
import logging

__version__ = '0.0.6'

try:
    # not available in python 2.6
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):

        def emit(self, record):
            pass

# Add NullHandler to prevent logging warnings
logging.getLogger(__name__).addHandler(NullHandler())


class TaskNotRegistered(Exception):
    '''It can be raised by the worker when it tries to execute a
    which there is not callback
    '''
    pass


class TaskAlreadyRegistered(Exception):
    '''It is raised when @task decorator or Kamma.add_task is being called with a task id already used.
    '''
    pass


class AbortTask(Exception):
    ''' Exception which can be raised by the callback in order to
    abort the executing task due to a decision taken by the application
    '''
    pass


class TimeoutTask(Exception):
    ''' Exception which can be raised internally in kamma when the
    callback lasts more time than its given 'timeout'
    '''
    pass


class RetryStopped(Exception):
    ''' When a task cannot be executed within limited retry policy
    '''

    def __init__(self, callback, attempts, delay, last_error):
        self.callback = callback
        self.attempts = attempts
        self.delay = delay
        self.last_error = last_error
        self.json_task = ""
        msg = "The task '{cb}' has failed after {tries} attempts and {sec} seconds. "\
            "Last Error: {err}".format(cb=callback, tries=attempts, sec=delay, err=last_error)
        super(RetryStopped, self).__init__(msg)
